/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Analyzer.h"
#include "Debugger.h"
#include "State.h"
#include "Util.h"
#include "MD5.h"
#include <QMenu>
#include <QProgressDialog>
#include <QCoreApplication>
#include "DebuggerCoreInterface.h"
#include "ArchProcessorInterface.h"
#include "SymbolManager.h"
#include <cstring>
#include <elf.h>

#define MIN_REFCOUNT 2

#if defined(EDB_X86)
static const edb::Operand::Register STACK_REG = edb::Operand::REG_ESP;
static const edb::Operand::Register FRAME_REG = edb::Operand::REG_EBP;
#elif defined(EDB_X86_64)
static const edb::Operand::Register STACK_REG = edb::Operand::REG_RSP;
static const edb::Operand::Register FRAME_REG = edb::Operand::REG_RBP;
#endif

//------------------------------------------------------------------------------
// Name: Analyzer(
// Desc: 
//------------------------------------------------------------------------------
Analyzer::Analyzer() : m_Menu(0) {
}


//------------------------------------------------------------------------------
// Name: menu(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
QMenu *Analyzer::menu(QWidget *parent) {

	edb::v1::setAnalyzer(this);

	if(m_Menu == 0) {
		m_Menu = new QMenu("Analyzer", parent);
		m_Menu->addAction ("&Analyzer Current Region", this, SLOT(doMenu()), QKeySequence("Ctrl+A"));
	}

	return m_Menu;
}

//------------------------------------------------------------------------------
// Name: doMenu()
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::doMenu() {
	MemoryRegions::Region region;
	const State &state = edb::v1::currentState();
	const edb::address_t eip = state.instructionPointer();
	if(edb::v1::memoryRegions().findRegion(eip, region)) {
	
		QProgressDialog *const dialog = new QProgressDialog("Performing Analysis", 0, 0, 100, edb::v1::debuggerUI);
		connect(this, SIGNAL(updatePercentage(int)), dialog, SLOT(setValue(int)));
		
		dialog->show();
		QCoreApplication::processEvents();
	
		analyze(region);
		delete dialog;
		
		edb::v1::repaintCPUView();
	}
}

//------------------------------------------------------------------------------
// Name: findFunctionCalls(const MemoryRegions::Region &region, FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::findFunctionCalls(const MemoryRegions::Region &region, FunctionMap &results) {
	static const edb::address_t pageSize = edb::v1::debuggerBase->pageSize();

	const edb::address_t size_in_pages = region.size() / pageSize;
	quint8 *pages = new quint8[size_in_pages * pageSize];

	if(edb::v1::debuggerBase->readPages(region.start, pages, size_in_pages)) {
		for(edb::address_t i = 0; i < size_in_pages * pageSize; ++i) {
			const edb::Instruction insn(&pages[i], region.size() - i, region.start + i, std::nothrow);

			if(insn.valid() && insn.type() == edb::Instruction::OP_CALL) {

				State state;
				state.setInstructionPointer(region.start + i);

				const edb::Operand &op = insn.operand(0);

				if(op.generalType() == edb::Operand::TYPE_REL) {
					edb::address_t ea = op.relativeTarget();

					// skip over ones which are : call <label>; label:
					if(ea != state.instructionPointer() + insn.size()) {
						if(ea >= region.start && ea < region.end) {

							results[ea].entry_address	= ea;
							results[ea].end_address		= ea;
							results[ea].reference_count++;	
						}
					}
				}

				emit updatePercentage(CalculatePercentage(0, 2, i, region.size()));
			}
			QCoreApplication::processEvents();
		}
	}

	release_array(pages);
}

//------------------------------------------------------------------------------
// Name: findStackFrames(const MemoryRegions::Region &region, FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::findStackFrames(const MemoryRegions::Region &region, FunctionMap &results) {

	for(edb::address_t address = region.start; address != region.end; ++address) {
		if(isStackFrame(address)) {
			results[address].entry_address		= address;
			results[address].end_address		= address;
			results[address].reference_count++;
		}
		
		emit updatePercentage(CalculatePercentage(1, 2, address - region.start, region.size()));
	}
}

//------------------------------------------------------------------------------
// Name: isStackFrame(edb::address_t addr) const
// Desc: 
//------------------------------------------------------------------------------
bool Analyzer::isStackFrame(edb::address_t addr) const {

	// TODO: make a version which can operate on memory
	// directly, so that we can prefetch whole pages

	quint8 buf[edb::Instruction::MAX_SIZE];

	unsigned int i = 0;

	while(i < 2) {
		// gets the bytes for the instruction
		int buf_size = sizeof(buf);
		if(!edb::v1::getInstructionBytes(addr, buf, buf_size)) {
			break;
		}

		// decode it
		const edb::Instruction insn(buf, buf_size, addr, std::nothrow);
		if(!insn.valid()) {
			break;;
		}

		// which part of the sequence are we looking at?
		switch(i++) {
		case 0:
			// are we looking at a "push rBP" ?
			if(insn.type() == edb::Instruction::OP_PUSH) {
				const edb::Operand &op = insn.operand(0);
				if(op.completeType() == edb::Operand::TYPE_REGISTER) {
					if(op.reg == FRAME_REG) {
						break;
					}
				}
			}
			++i;
			break;
		case 1:
			// are we looking at a "mov rBP, rSP" ?
			if(insn.type() == edb::Instruction::OP_MOV) {
				const edb::Operand &op0 = insn.operand(0);
				const edb::Operand &op1 = insn.operand(1);
				if(op0.completeType() == edb::Operand::TYPE_REGISTER && op1.completeType() == edb::Operand::TYPE_REGISTER) {
					if(op0.reg == FRAME_REG && op1.reg == STACK_REG) {
						return true;
					}
				}
			}
			break;
		default:
			break;
		}

		addr += insn.size();

		QCoreApplication::processEvents();
	}
	
	return false;
}

//------------------------------------------------------------------------------
// Name: bonusStackFrames(FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonusStackFrames(FunctionMap &results) {
	// give bonus if we see a:
	// push ebp; mov ebp, esp;

	for(FunctionMap::iterator it = results.begin(); it != results.end(); ++it) {
		
		Function &info = it.value();
		const edb::address_t addr = info.entry_address;
		
		const bool bonus = isStackFrame(addr);

		
		if(bonus) {
			info.reference_count++;
		}
	}
}

//------------------------------------------------------------------------------
// Name: bonusMain(const MemoryRegions::Region &region, FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonusMain(const MemoryRegions::Region &region, FunctionMap &results) {
	const QString s = edb::v1::getProcessExecutable();
	if(!s.isEmpty()) {
		const edb::address_t main = edb::v1::locateMainFunction(s);
		
		if(main != 0x00000000) {
		
			if(main >= region.start && main < region.end) {
				// make sure we have an entry for this function
				results[main].entry_address		= main;
				results[main].end_address		= main;
				if(results[main].reference_count == 0) {
					results[main].reference_count = 2;
				} else {
					results[main].reference_count++;
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: bonusSymbols(FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonusSymbols(FunctionMap &results) {
	// give bonus if we have a symbol for the address

	for(FunctionMap::iterator it = results.begin(); it != results.end(); ++it) {
		const edb::address_t addr = it.value().entry_address;

		const SymbolManager::Symbol *const sym = edb::v1::symbolManager().findSymbol(addr);
		if(sym != 0) {

			results[addr].entry_address		= addr;
			results[addr].end_address		= addr;
			results[addr].reference_count++;
		}
		
		QCoreApplication::processEvents();
	}
}

//------------------------------------------------------------------------------
// Name: findFunctionEnds(FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
int Analyzer::findFunctionEnds(FunctionMap &results) {
	int ret = 0;
	FunctionMap::iterator it = results.begin();
	while(it != results.end()) {
		Function &info = it.value();
		
		const FunctionMap::iterator next = ++it;
		
		if(info.reference_count >= MIN_REFCOUNT) {
			int updates;
			edb::address_t next_entry;
		
			if(next != results.end()) {
				next_entry = next.value().entry_address;
			} else {
				// just round up to next page, since we don't have a good guess for end...
				next_entry = ((info.entry_address + (0x1000 - 1)) & ~(0x1000 - 1));
				
			}
			edb::address_t last_instruction;
			const edb::address_t end = findFunctionEnd(info.entry_address, next_entry, updates, results, last_instruction);
			ret += updates;
			info.end_address		= end;
			info.last_instruction	= last_instruction;


			quint8 buf[edb::Instruction::MAX_SIZE];
			int buf_size = sizeof(buf);
			if(edb::v1::getInstructionBytes(last_instruction, buf, buf_size)) {
				const edb::Instruction insn(buf, buf_size, last_instruction, std::nothrow);
				if(insn.valid() && insn.type() == edb::Instruction::OP_JMP) {
				
					const edb::Operand &op = insn.operand(0);
					
					if(op.generalType() == edb::Operand::TYPE_REL) {
						const edb::address_t target = op.relativeTarget();
						
						Function func;
						
						if(!findContainingFunction(target, func)) {
						
							results[target].entry_address	= target;
							results[target].end_address		= target;
							results[target].reference_count = 2;
							++ret;
						}
					}
				}
			}
		}	
		
		it = next;
		QCoreApplication::processEvents();
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: findFunctionEnd(edb::address_t start_address, edb::address_t end_address, int &update_count, FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
edb::address_t Analyzer::findFunctionEnd(edb::address_t start_address, edb::address_t end_address, int &update_count, FunctionMap &results, edb::address_t &last_instruction) {
	edb::address_t				ret = start_address;
	QList<edb::address_t>		jump_targets;
	QMap<edb::address_t, int>	code_addresses;
	
	update_count = 0;
	
	// we start with the entry point of the function
	jump_targets.push_front(start_address);
	
	// while no more jump targets... (includes entry point)
	while(!jump_targets.empty()) {
		edb::address_t addr = jump_targets.front();
		jump_targets.pop_front();

		// keep going until we go out of bounds	
		while(addr >= start_address && addr < end_address) {
				
			quint8 buf[edb::Instruction::MAX_SIZE];
			int buf_size = sizeof(buf);
			if(!edb::v1::getInstructionBytes(addr, buf, buf_size)) {
				break;
			}
			
			const edb::Instruction insn(buf, buf_size, addr, std::nothrow);
			if(!insn.valid()) {
				break;
			}

			// ok, it was a valid instruction, let's add it to the
			// list of possible "code addresses"
			code_addresses.insert(addr, insn.size());
			
			const edb::Instruction::Type type = insn.type();

			// is it a branch?
			if(type == edb::Instruction::OP_JCC) {

				const edb::Operand &op = insn.operand(0);
				if(op.generalType() == edb::Operand::TYPE_REL) {
					const edb::address_t ea = op.relativeTarget();

					if(!code_addresses.contains(ea)) {
						if(!jump_targets.contains(ea)) {
							jump_targets.push_front(ea);
						}
					}
				}
			} else if(type == edb::Instruction::OP_JMP) {

				const edb::Operand &op = insn.operand(0);
				if(op.generalType() == edb::Operand::TYPE_REL) {
					const edb::address_t ea = op.relativeTarget();
					
					// is it a jump to another function's entry point?
					// if so, this is a dead end, resume from other branches
					// but give the target a bonus reference
					FunctionMap::iterator it = results.find(ea);
					if(it != results.end()) {
						if(it.value().reference_count < MIN_REFCOUNT) {
							it.value().reference_count++;
							++update_count;
							break;
						}
					}
					
					
					if(!code_addresses.contains(ea)) {
						addr = ea;
						continue;
					} else {
						break;
					}
				}
			} else if(type == edb::Instruction::OP_CALL) {

				const edb::Operand &op = insn.operand(0);
				if(op.generalType() == edb::Operand::TYPE_REL) {
					const edb::address_t ea = op.relativeTarget();

					// skip over ones which are : call <label>; label:
					if(ea != addr + insn.size()) {
					
						FunctionMap::iterator it = results.find(ea);
						if(it != results.end()) {
							if(it.value().reference_count < MIN_REFCOUNT) {
								it.value().reference_count++;
								++update_count;
							}
						}					
					}
				} else if(op.generalType() == edb::Operand::TYPE_EXPRESSION) {
				#if 0
					printf("found jump table? %08x\n", addr);
				#endif
				}
			} else if(type == edb::Instruction::OP_RET) {
				break;
			} else if(type == edb::Instruction::OP_HLT) {
				break;
			}
			
			addr += insn.size();
			QCoreApplication::processEvents();
		}
	}
	
	last_instruction = 0;
	
	// find the last entry
	for(QMap<edb::address_t, int>::const_iterator it = code_addresses.begin(); it != code_addresses.end(); ++it) {
		const edb::address_t a = it.key() + it.value() - 1;
		ret = qMax(a, ret);
		QCoreApplication::processEvents();
		
		// update entry for last instruction, (useful for detecting call;ret -> jmp optimization)
		last_instruction = qMax(last_instruction, it.key());
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: isThunk(edb::address_t address) const
// Desc: basically returns true if the first instruction of the function is a 
//       jmp
//------------------------------------------------------------------------------
bool Analyzer::isThunk(edb::address_t address) const {
	quint8 buf[edb::Instruction::MAX_SIZE];
	int buf_size = sizeof(buf);
	bool ret = false;
	if(edb::v1::getInstructionBytes(address, buf, buf_size)) {
		const edb::Instruction insn(buf, buf_size, address, std::nothrow);
		ret = insn.valid() && insn.type() == edb::Instruction::OP_JMP;
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: setFunctionTypes(FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::setFunctionTypes(FunctionMap &results) {
	// give bonus if we have a symbol for the address

	for(FunctionMap::iterator it = results.begin(); it != results.end(); ++it) {
		const edb::address_t addr = it.value().entry_address;

		if(isThunk(addr)) {
			it.value().type = Function::FUNCTION_THUNK;
		} else {
			it.value().type = Function::FUNCTION_STANDARD;
		}
		
		QCoreApplication::processEvents();
	}
}

//------------------------------------------------------------------------------
// Name: analyze(const MemoryRegions::Region &region)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::analyze(const MemoryRegions::Region &region) {

	const QByteArray md5		= md5Region(region);
	const QByteArray prev_md5	= m_AnalysisMD5[region];
	
	if(md5 != prev_md5) {
		FunctionMap &results = m_AnalysisResults[region];

		results.clear();

		edb::v1::log("[Analyzer] finding obvious function calls...");
		findFunctionCalls(region, results);

		edb::v1::log("[Analyzer] finding obvious stack frames...");
		findStackFrames(region, results);

		edb::v1::log("[Analyzer] adding ELF entry points to the list...");
		bonusELFEntryPoint(region, results);

		edb::v1::log("[Analyzer] attempting to add 'main' to the list...");
		bonusMain(region, results);

		edb::v1::log("[Analyzer] attempting to add functions with symbols to the list...");
		bonusSymbols(results);

		edb::v1::log("[Analyzer] bonusing stack frames...");
		bonusStackFrames(results);

		int updates;
		do {
			edb::v1::log("[Analyzer] calculating function bounds...");
			updates = findFunctionEnds(results);
			edb::v1::log("[Analyzer] got %d updates", updates);
		} while(updates != 0);

		edb::v1::log("[Analyzer] determining function types...");
		setFunctionTypes(results);

		edb::v1::log("[Analyzer] complete");

		m_AnalysisMD5[region]= md5;
	} else {
		edb::v1::log("[Analyzer] region unchanged, using previous analysis");
	}
}

//------------------------------------------------------------------------------
// Name: catagory(edb::address_t address) const
// Desc: 
//------------------------------------------------------------------------------
AnalyzerInterface::AddressCatagory Analyzer::catagory(edb::address_t address) const {

	Function func;
	if(findContainingFunction(address, func)) {
		if(address == func.entry_address) {
			return ADDRESS_FUNC_START;
		} else if(address == func.end_address) {
			return ADDRESS_FUNC_END;
		} else {
			return ADDRESS_FUNC_BODY;
		}
	}
	return ADDRESS_FUNC_UNKNOWN;
}

//------------------------------------------------------------------------------
// Name: functions(const MemoryRegions::Region &region) const
// Desc: 
//------------------------------------------------------------------------------
AnalyzerInterface::FunctionMap Analyzer::functions(const MemoryRegions::Region &region) const {

	const FunctionMap &results = m_AnalysisResults[region];
	return results;
}

//------------------------------------------------------------------------------
// Name: findContainingFunction(edb::address_t address, AnalyzerInterface::Function &function) const
// Desc: 
//------------------------------------------------------------------------------
bool Analyzer::findContainingFunction(edb::address_t address, AnalyzerInterface::Function &function) const {

	MemoryRegions::Region region;
	if(edb::v1::memoryRegions().findRegion(address, region)) {
		const FunctionMap &funcs = functions(region);
		foreach(Function f, funcs) {
			if(address >= f.entry_address && address <= f.end_address) {
				function = f;
				return true;
			}
		}
	}
	return false;
}


//------------------------------------------------------------------------------
// Name: md5Region(const MemoryRegions::Region &region)
// Desc: returns a byte array representing the MD5 of a region
//------------------------------------------------------------------------------
QByteArray Analyzer::md5Region(const MemoryRegions::Region &region) {

	static const edb::address_t pageSize = edb::v1::debuggerBase->pageSize();

	MD5 md5Digest;
	QByteArray ret;

	const edb::address_t size_in_pages = region.size() / pageSize;
	quint8 *pages = new quint8[size_in_pages * pageSize];

	if(edb::v1::debuggerBase->readPages(region.start, pages, size_in_pages)) {
		md5Digest.update(pages, size_in_pages * pageSize);
		md5Digest.finalize();
		
		// make a deep copy because MD5 is about to go out of scope
		ret = QByteArray(reinterpret_cast<const char *>(md5Digest.digest()), 16);
	}

	release_array(pages);
	
	return ret;
}


//------------------------------------------------------------------------------
// Name: bonusELFEntryPoint(const MemoryRegions::Region &region, FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonusELFEntryPoint(const MemoryRegions::Region &region, FunctionMap &results) {
	static const edb::address_t pageSize = edb::v1::debuggerBase->pageSize();

	union {
		Elf32_Ehdr	*elf32;
		Elf64_Ehdr	*elf64;
		quint8		*pages;
	} code;
	
	edb::address_t entry = 0x00000000;
	
	code.pages = new quint8[pageSize];

	if(edb::v1::debuggerBase->readPages(region.start, code.pages, 1)) {
		
		if(std::memcmp(code.elf32->e_ident, ELFMAG, SELFMAG) == 0) {
			if(code.elf32->e_ident[EI_CLASS] == ELFCLASS32) {
				entry = code.elf32->e_entry;
			}
		}
		
		if(entry == 0x00000000 && std::memcmp(code.elf64->e_ident, ELFMAG, SELFMAG) == 0) {
			if(code.elf64->e_ident[EI_CLASS] == ELFCLASS64) {
				entry = code.elf64->e_entry;
			}
		}
	}

	release_array(code.pages);

	if(entry != 0x00000000) {
	
		// if the entry seems like a relative one (like for a library)
		// then add the base of its image
		if(entry < region.start) {
			entry += region.start;
		}
	
		// make sure we have an entry for this function
		
		if(entry >= region.start && entry < region.end) {		
			results[entry].entry_address	= entry;
			results[entry].end_address		= entry;
			if(results[entry].reference_count == 0) {
				results[entry].reference_count = 2;
			} else {
				results[entry].reference_count++;
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: invalidateAnalysis(const MemoryRegions::Region &region)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::invalidateAnalysis(const MemoryRegions::Region &region) {

	m_AnalysisResults[region].clear();
	m_AnalysisMD5[region].clear();
}

//------------------------------------------------------------------------------
// Name: invalidateAnalysis()
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::invalidateAnalysis() {
	m_AnalysisResults.clear();
	m_AnalysisMD5.clear();
}

Q_EXPORT_PLUGIN2(Analyzer, Analyzer)
